% This script demonstrates the use of the code to solve maximum
% likelihood estimation of the scale parameter ell in the Matern
% covariance kernel for the following configuration of the points:
%
%   2D, no grid structure.
%
% This script requires defining an auxiliary function
%
%   ex6_aux_func
%
% to appropriately handle matrix-vector multiplications and a wrapper
%
%   ex6_score_func
%
% to incorporate these multiplications.
%

%Author: J. Chen
%$Date: 2011/12/20 12:49:48$

clear
addpath(genpath('../src'))


%---------------- PARAMETERS ----------------
n = 1024;           % number of points
d = 2;              % dimension
T = [100;100];      % physical dimension

nu = 2;             % Matern: smoothness
ell = [7;10];       % Matern: scale

ell0 = [10;7];      % nonlinear solver: initial guess
N = 50;             % nonlinear solver: # samples in SAA
tol_fsolve = 1e-4;  % nonlinear solver: tol

m = 100;            % linear solver: deflation dimension
maxit = 200;        % linear solver: maxit
tol_cg = 1e-8;      % linear solver: tol

N2 = 100;           % Fisher trace approx: # samples in SAA

RandStream.setDefaultStream(RandStream('mt19937ar','seed',1420));


%---------------- GENERATE GP SAMPLE ----------------
fprintf(1, '--> Generating GP sample ...\n');

% generate random points
[coord, tri] = gen_coord_random_2d(n, T);

% compute boundary of the points
boundary = boundary2d(tri);

% generate covariance matrix based on true ell
K = gen_K_matern_general(coord, boundary, ell, nu);

% sample GP using the Cholesky method
y = sampl_chol(K);


%---------------- COMPUTE PRECONDITIONER ----------------
fprintf(1, '--> Computing preconditioner ...\n');

% compute stiffness matrix preconditioner
M = stiff2d(coord, tri, boundary);
tau = round(nu+d/2);


%---------------- GENERATE RADEMACHER VECTORS ----------------
fprintf(1, '--> Generating Rademacher vectors ...\n');

% generate N independent Rademacher vectors
u = sign(randn(length(y), N));
u(u==0) = 1;


%---------------- SOLVE MAX LOG-LIKELIHOOD ----------------
fprintf(1, '--> Solving max likelihood ...\n');

% solve the SAA of the score equation
% use deflation in inner linear solver (m ~= 0)
tic;
[ell_est, fval, exitflag, output, jacob] = ...
    fsolve(@(ell)ex6_score_func(coord, boundary, ell, nu, M, tau, ...
                                y, u, m, maxit, tol_cg), ...
           ell0, optimset('Display','iter','TolX',tol_fsolve));
elapsed_time = toc;


%---------------- COMPUTE COVARIANCE OF SAA ----------------
fprintf(1, '--> Computing covariance of SAA ...\n');

% compute covariance of SAA
[Kfun, dKfun, Mfun] = ex6_aux_func(coord, boundary, ell_est, nu, M, tau);
Gamma = score_func_saa_covar(Kfun, dKfun, Mfun, ...
                             y, u, m, maxit, tol_cg, jacob);


%---------------- COMPUTE COVARIANCE OF MLE ----------------
fprintf(1, '--> Computing covariance of MLE ...\n');
%{
% compute covariance of MLE
[K, dK] = gen_K_matern_general(coord, boundary, ell, nu);
[I0, V0] = score_func_mle_covar(K, dK, N);
%}
% compute covariance of MLE (approximately)
u2 = sign(randn(length(y), N2));
u2(u2==0) = 1;

[Kfun, dKfun, Mfun] = ex6_aux_func(coord, boundary, ell, nu, M, tau);
[I, V] = score_func_mle_covar_approx(Kfun, dKfun, Mfun, N, ...
                                     u2, m, maxit, tol_cg);


%---------------- PRINT RESULTS ----------------
fprintf(1, '--> Done.\n\n');
fprintf(1, 'MAX LIKELIHOOD RESULTS:\n');
fprintf(1, 'Truth:      ell = %8.5f, %8.5f\n', ell);
fprintf(1, 'Init guess: ell = %8.5f, %8.5f\n', ell0);
fprintf(1, 'Estimate:   ell = %8.5f, %8.5f\n', ell_est);
fprintf(1, 'Walltime:   %g sec\n', elapsed_time);
%{
fprintf(1, 'MLE covariance (inverse of Fisher matrix I):\n\n');
invI0 = inv(I0);
disp(invI0);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(invI0));

fprintf(1, 'MLE covariance with SAA (inv(A)*B*inv(A)):\n\n');
disp(V0);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(V0));
%}
fprintf(1, 'MLE covariance (approximately):\n\n');
invI = inv(I);
disp(invI);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(invI));

fprintf(1, 'MLE covariance with SAA (approximately):\n\n');
disp(V);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(V));

fprintf(1, 'SAA covariance (inv(jacob)''*V*inv(jacob)/N):\n\n');
disp(Gamma);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(Gamma));
