% This script demonstrates the use of the code to solve maximum
% likelihood estimation of the scale parameter ell in the power law
% generalized covarinace kernel for the following configuration of the
% points:
%
%   2D, regular grid, containing missing regions in disc shape.
%
% This script requires defining an auxiliary function
%
%   ex9_aux_func
%
% to appropriately handle matrix-vector multiplications and a wrapper
%
%   ex9_score_func
%
% to incorporate these multiplications.
%

%Author: J. Chen
%$Date: 2011/12/20 12:50:11$

clear
addpath(genpath('../src'))


%---------------- PARAMETERS ----------------
dim = [32;32];         % grid: size
T = [100;100];         % grid: physical dimension
discs = {[40;60;20]};  % grid: discs ([x-coord, y-coord, radius])

alpha = 1.5;           % kernel parameter
ell = [7;10];          % kernel parameter

ell0 = [10;7];         % nonlinear solver: initial guess
N = 50;                % nonlinear solver: # samples in SAA
tol_fsolve = 1e-4;     % nonlinear solver: tol

m = 0;                 % linear solver: deflation dimension
maxit = 200;           % linear solver: maxit
tol_cg = 1e-6;         % linear solver: tol

N2 = 100;              % Fisher trace approx: # samples in SAA

num_filter = round((alpha+length(dim))/4);  % filtering

RandStream.setDefaultStream(RandStream('mt19937ar','seed',1420));


%---------------- GENERATE GP SAMPLE ----------------
fprintf(1, '--> Generating GP sample ...\n');

% generate covariance matrix for regular grid based on current ell
tK = gen_K_power_regular(dim, T, ell, alpha);

% filter K using Laplace filter
% this has to be done to ensure positive definiteness of K
tK = lapFilterMat2d(tK, num_filter);

% sample GP using the embedding method
[y, inc] = sampl_embed(tK, 4);

% find points corresponding to the missing regions
invalid_pts = find_invalid_pts(dim, T, discs, num_filter);

% take care of y
y(invalid_pts) = [];


%---------------- GENERATE RADEMACHER VECTORS ----------------
fprintf(1, '--> Generating Rademacher vectors ...\n');

% generate N independent Rademacher vectors
n = length(y);
u = sign(randn(n, N));
u(u==0) = 1;


%---------------- SOLVE MAX LOG-LIKELIHOOD ----------------
fprintf(1, '--> Solving max likelihood ...\n');

% solve the SAA of the score equation
% do not use deflation in inner linear solver (m = 0)
tic;
[ell_est, fval, exitflag, output, jacob] = ...
    fsolve(@(ell)ex9_score_func(dim, T, ell, alpha, num_filter, ...
                                invalid_pts, y, u, m, maxit, tol_cg), ...
           ell0, optimset('Display','iter','TolX',tol_fsolve));
elapsed_time = toc;


%---------------- COMPUTE COVARIANCE OF SAA ----------------
fprintf(1, '--> Computing covariance of SAA ...\n');

% compute covariance of SAA
[Kfun, dKfun, Mfun] = ex9_aux_func(dim, T, ell_est, alpha, ...
                                   num_filter, invalid_pts, 'func');
Gamma = score_func_saa_covar(Kfun, dKfun, Mfun, ...
                             y, u, m, maxit, tol_cg, jacob);


%---------------- COMPUTE COVARIANCE OF MLE ----------------
fprintf(1, '--> Computing covariance of MLE ...\n');
%{
% compute covariance of MLE
[K, dK] = ex9_aux_func(dim, T, ell, alpha, num_filter, invalid_pts, 'mat');
[I0, V0] = score_func_mle_covar(K, dK, N);
%}
% compute covariance of MLE (approximately)
u2 = sign(randn(n, N2));
u2(u2==0) = 1;

[Kfun, dKfun, Mfun] = ex9_aux_func(dim, T, ell, alpha, ...
                                   num_filter, invalid_pts, 'func');
[I, V] = score_func_mle_covar_approx(Kfun, dKfun, Mfun, N, ...
                                     u2, m, maxit, tol_cg);


%---------------- PRINT RESULTS ----------------
fprintf(1, '--> Done.\n\n');
fprintf(1, 'MAX LIKELIHOOD RESULTS:\n');
fprintf(1, 'Truth:      ell = %8.5f, %8.5f\n', ell);
fprintf(1, 'Init guess: ell = %8.5f, %8.5f\n', ell0);
fprintf(1, 'Estimate:   ell = %8.5f, %8.5f\n', ell_est);
fprintf(1, 'Walltime:   %g sec\n', elapsed_time);
%{
fprintf(1, 'MLE covariance (inverse of Fisher matrix I):\n\n');
invI0 = inv(I0);
disp(invI0);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(invI0));

fprintf(1, 'MLE covariance with SAA (inv(A)*B*inv(A)):\n\n');
disp(V0);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(V0));
%}
fprintf(1, 'MLE covariance (approximately):\n\n');
invI = inv(I);
disp(invI);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(invI));

fprintf(1, 'MLE covariance with SAA (approximately):\n\n');
disp(V);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(V));

fprintf(1, 'SAA covariance (inv(jacob)''*V*inv(jacob)/N):\n\n');
disp(Gamma);

fprintf(1, 'eigenvalues:\n\n');
disp(eig(Gamma));
