function [idx] = find_invalid_pts(dim, T, discs, num_filter)
%function [idx] = find_invalid_pts(dim, T, discs, num_filter)
%
% This function finds a set A of points in a grid that are inside the
% specified discs. Points that are not inside A but are in the filter
% range of A (thus we call them "boundary of A") are also
% included. Because of filtering, a number of layers (specified by
% num_filter) on the boundary of the grid become useless, thus the
% indices of the points in A (including boundary of A) are shifted by
% num_filter along each dimension. In other words, the functionality
% of this routine is such that if K is the filtered covariance matrix
% for the grid, then by deleting the rows and columns specified by
% idx, the resulting K becomes the filtered covariance matrix for only
% the points that are outside the disc regions.
%
% Only grids up to 3D are supported.
%
% INPUTS
%   dim:        Grid size.
%   T:          Physical size of the grid.
%   discs:      Disc regions. It is a cell array, with each cell
%               denoting a disc. E.g., if discs{1} = [40;60;20],
%               then it means a disc centered at (40,60), with
%               radius 20.
%   num_filter: Number of filtering.
%
% OUTPUTS
%   idx:        Indices of the points in A and boundary of A.
%

%Author: J. Chen
%$Date: 2011/12/08 14:21:21$

d = length(dim);
if d > 3
  error('Only grids up to 3D are supported');
end

if length(T) == 1
  T = repmat(T,1,d);
end

for i = 1:d
  co{i} = linspace(0, T(i), dim(i))';
end

if d == 1
  CO{1} = co{1};
elseif d == 2
  [CO{1}, CO{2}] = ndgrid(co{1}, co{2});
elseif d == 3
  [CO{1}, CO{2}, CO{3}] = ndgrid(co{1}, co{2}, co{3});
end

% get points in the discs
idx = [];
sz = size(CO{1});
for j = 1:length(discs)
  dd = zeros(sz);
  for i = 1:d
    dd = dd + (CO{i} - discs{j}(i)).^2;
  end
  idx = [idx; find(dd <= discs{j}(d+1)^2)];
end

if d == 1

  % include points on the boundary of the discs
  i1 = idx;
  len = length(i1);
  i1 = [repmat(i1,1,2*num_filter+1) + repmat(-num_filter:num_filter,len,1)];

  % shift indices since boundary of the grid is not used due to filtering
  i1 = i1 - num_filter;
  
  % trim ghost points
  ind = find(i1<=0 | i1>dim(1)-2*num_filter);
  i1(ind) = [];
  
  % finally get the linear indices
  idx = i1;

elseif d == 2

  % include points on the boundary of the discs
  [i1, i2] = ind2sub(sz, idx);
  len = length(i1);
  i1 = [repmat(i1,1,2*num_filter+1) + repmat(-num_filter:num_filter,len,1) ...
        repmat(i1,1,2*num_filter+1)];
  i2 = [repmat(i2,1,2*num_filter+1) ...
        repmat(i2,1,2*num_filter+1) + repmat(-num_filter:num_filter,len,1)];

  % shift indices since boundary of the grid is not used due to filtering
  i1 = i1 - num_filter;
  i2 = i2 - num_filter;
  
  % trim ghost points
  i1 = i1(:);
  i2 = i2(:);
  ind = find(i1<=0 | i1>dim(1)-2*num_filter | ...
             i2<=0 | i2>dim(2)-2*num_filter);
  i1(ind) = [];
  i2(ind) = [];
  
  % finally get the linear indices
  idx = sub2ind(sz-2*num_filter, i1, i2);
  
elseif d == 3

  % include points on the boundary of the discs
  [i1, i2, i3] = ind2sub(sz, idx);
  len = length(i1);
  i1 = [repmat(i1,1,2*num_filter+1) + repmat(-num_filter:num_filter,len,1) ...
        repmat(i1,1,2*num_filter+1) ...
        repmat(i1,1,2*num_filter+1)];
  i2 = [repmat(i2,1,2*num_filter+1) ...
        repmat(i2,1,2*num_filter+1) + repmat(-num_filter:num_filter,len,1) ...
        repmat(i2,1,2*num_filter+1)];
  i3 = [repmat(i3,1,2*num_filter+1) ...
        repmat(i3,1,2*num_filter+1) ...
        repmat(i3,1,2*num_filter+1) + repmat(-num_filter:num_filter,len,1)];
    
  % shift indices since boundary of the grid is not used due to filtering
  i1 = i1 - num_filter;
  i2 = i2 - num_filter;
  i3 = i3 - num_filter;
  
  % trim ghost points
  i1 = i1(:);
  i2 = i2(:);
  i3 = i3(:);
  ind = find(i1<=0 | i1>dim(1)-2*num_filter | ...
             i2<=0 | i2>dim(2)-2*num_filter | ...
             i3<=0 | i3>dim(3)-2*num_filter);
  i1(ind) = [];
  i2(ind) = [];
  i3(ind) = [];
  
  % finally get the linear indices
  idx = sub2ind(sz-2*num_filter, i1, i2, i3);
  
end

% remove repetitions
idx = unique(idx);
