function [K, dK] = gen_K_matern_general(coord, boundary, ell, nu)
%function [K, dK] = gen_K_matern_general(coord, boundary, ell, nu)
%
% This function generates a covariance matrix (based on the Matern
% kernel) and its derivatives, for a set of points not necessarily
% on a regular grid.
%
% INPUTS
%   coord:    Coordinates of the points. Each row represents one
%             point.
%   boundary: Indices of the points that serve as boundary. These
%             points are excluded in the covariance matrices. One
%             can input an empty vector if one does not want to
%             exclude points.
%   ell:      Matern scale parameter. E.g., [7;13].
%   nu:       Matern smoothness parameter. E.g., 1.5.
%
% OUTPUTS
%   K:        Covariance matrix.
%   dK:       A cell array containing the partial derivatives of
%             the covariance matrix w.r.t. to ell.
%

%Author: J. Chen
%$Date: 2011/12/07 13:17:15$

[n,d] = size(coord);

r = zeros(n);
for i = 1:d
  dist{i} = repmat( coord(:,i),1,n ) - repmat( coord(:,i)',n,1 );
  dist{i} = dist{i}.^2/ell(i).^2;
  r = r + dist{i};
end
r = sqrt(2*nu*r);

% K
K = r.^nu .* besselk(nu,r);
K0 = 2^(nu-1)*gamma(nu);
for j = 1:n
  K(j,j) = K0;
end
K = K/K0;
K(:,boundary) = [];
K(boundary,:) = [];

% derivatives
if nargout > 1
  bes0 = 2*nu/K0 * r.^(nu-1) .* besselk(nu-1,r);
  for j = 1:n
    bes0(j,j) = 0;
  end
  for i = 1:d
    dK{i} = bes0 .* dist{i} / ell(i);
    dK{i}(:,boundary) = [];
    dK{i}(boundary,:) = [];
  end
end
