function [tK, tdK] = gen_K_matern_regular(dim, T, ell, nu)
%function [tK, tdK] = gen_K_matern_regular(dim, T, ell, nu)
%
% This function generates a covariance matrix (based on the Matern
% kernel) and its derivatives, for a set of points on a regular
% grid. Only grids up to 3D are supported.
%
% INPUTS
%   dim: Grid size. E.g., [100;100] denotes a 100x100 grid.
%   T:   Physical size of the grid. E.g., [50;50].
%   ell: Matern scale parameter. E.g., [7;13].
%   nu:  Matern smoothness parameter. E.g., 1.5.
%
% OUTPUTS
%   tK:  Covariance matrix. Returns the first column, reshaped to
%        the grid size.
%   tdK: Partial derivatives of the covariance matrix w.r.t. to
%        ell. It is a cell array, each element of which represents one
%        derivative. For each matrix only the first column is
%        returned, and the column is reshaped to the grid size.
%

%Author: J. Chen
%$Date: 2011/12/07 13:16:58$

d = length(dim);
if d > 3
  error('Only grids up to 3D are supported');
end

if length(T) == 1
  T = repmat(T,1,d);
end

for i = 1:d
  co{i} = linspace(0, T(i), dim(i))';
end

if d == 1
  CO{1} = co{1};
elseif d == 2
  [CO{1}, CO{2}] = ndgrid(co{1}, co{2});
elseif d == 3
  [CO{1}, CO{2}, CO{3}] = ndgrid(co{1}, co{2}, co{3});
end

r = zeros(size(CO{1}));
for i = 1:d
  CO{i} = CO{i}.^2/ell(i).^2;
  r = r + CO{i};
end
r = sqrt(2*nu*r);

% K
tK = r.^nu .* besselk(nu,r);
tK0 = 2^(nu-1)*gamma(nu);
tK(1) = tK0;
tK = tK/tK0;

% derivatives
if nargout > 1
  bes0 = 2*nu/tK0 * r.^(nu-1) .* besselk(nu-1,r);
  bes0(1) = 0;
  for i = 1:d
    tdK{i} = bes0 .* CO{i} / ell(i);
  end
end
