function [tK, tdK] = gen_K_power_regular(dim, T, ell, alpha)
%function [tK, tdK] = gen_K_power_regular(dim, T, ell, alpha)
%
% This function generates a covariance matrix (based on the power law
% generalized covariance kernel) and its derivatives, for a set of
% points on a regular grid. Only grids up to 3D are supported.
%
% This function requires the digamma function psi that can be
% evaluated at negative values.
%
% INPUTS
%   dim:   Grid size. E.g., [100;100] denotes a 100x100 grid.
%   T:     Physical size of the grid. E.g., [50;50].
%   ell:   Parameter. E.g., [7;13].
%   alpha: Parameter. E.g., 2.
%
% OUTPUTS
%   tK:    Covariance matrix. Returns the first column, reshaped to
%          the grid size.
%   tdK:   Partial derivatives of the covariance matrix w.r.t. to ell
%          and alpha.  It is a cell array, each element of which
%          represents one derivative. For each matrix only the first
%          column is returned, and the column is reshaped to the grid
%          size.
%

%Author: J. Chen
%$Date: 2011/12/19 16:59:16$

d = length(dim);
if d > 3
  error('Only grids up to 3D are supported');
end

if length(T) == 1
  T = repmat(T,1,d);
end

for i = 1:d
  co{i} = linspace(0, T(i), dim(i))';
end

if d == 1
  CO{1} = co{1};
elseif d == 2
  [CO{1}, CO{2}] = ndgrid(co{1}, co{2});
elseif d == 3
  [CO{1}, CO{2}, CO{3}] = ndgrid(co{1}, co{2}, co{3});
end

r = zeros(size(CO{1}));
for i = 1:d
  CO{i} = CO{i}.^2/ell(i).^2;
  r = r + CO{i};
end
r = sqrt(r);

if mod(alpha,2) ~= 0
  coef = gamma(-alpha/2);
else
  coef = (-1)^(1+alpha/2);
end

if mod(alpha,2) ~= 0
  
  % K
  tK = coef * r.^alpha;

  % derivatives
  if nargout > 1
    for i = 1:d
      tdK{i} = -coef*alpha * r.^(alpha-2) .* CO{i} / ell(i);
      tdK{i}(1) = 0;
    end
    tdK{d+1} = coef * r.^alpha .* log(r) + ...
        coef*psi(-alpha/2)*(-1/2) * r.^alpha;
    tdK{d+1}(1) = 0;
  end

else
  
  % K
  tK = coef * r.^alpha .* log(r);
  tK(1) = 0;

  % derivatives
  if nargout > 1
    for i = 1:d
      tdK{i} = -coef * r.^(alpha-2) .* (alpha*log(r)+1) .* CO{i} / ell(i);
      tdK{i}(1) = 0;
    end
    tdK{d+1} = NaN;
  end

end
