function [X, iter, res, res2, res3] = dbpcg(A, B, M, W, X0, maxit, tol, SOL)
%function [X, iter, res, res2, res3] = dbpcg(A, B, M, W, X0, maxit, tol, SOL)
%
% DBPCG: Deflated Block Preconditioned Conjugate Gradient
%
%                            M*AX = M*B
%
% INPUTS
%   A:     Matrix/Operator A.
%   B:     Right-hand sides.
%   M:     Preconditioner M. Can be empty.
%   W:     Deflation matrix.
%   X0:    Initial guesses.
%   maxit: Maximum number of iterations.
%   tol:   Tolerance of residual norm.
%   SOL:   (Optional) Actual solutions.
%
% OUTPUTS
%   X:     Computed solutions.
%   iter:  Actual number of iterations.
%   res:   Residual norm |b-Ax|/|b|, each system, all iter.
%   res2:  (If sol available) Error norm |x-sol|_A / |b|, each system, all iter.
%   res3:  (If sol available) Error norm |x-sol| / |sol|, each system, all iter.
%

%Author: J. Chen
%$Date: 2011/11/23 11:50:40$

if isnumeric(A)
  A = @(x)(A*x);
end
if ~isempty(M) & isnumeric(M)
  M = @(x)(M*x);
end

% compute something needed repeatedly
WW = W'*W;
AW = A(W);
WAW = W'*AW;

% normalize each linear system
X = X0;
num_vec = size(B,2);
normB = zeros(1,num_vec);
for i = 1:num_vec
  normB(i) = norm(B(:,i));
  B(:,i) = B(:,i) / normB(i);
  X(:,i) = X(:,i) / normB(i);
end
if exist('SOL', 'var')
  normSOL = zeros(1,num_vec);
  for i = 1:num_vec
    SOL(:,i) = SOL(:,i) / normB(i);
    normSOL(i) = norm(SOL(:,i));
  end
end

% allocate space to hold residuals
res = zeros(maxit,num_vec);
res2 = [];
res3 = [];
if exist('SOL', 'var')
  res2 = zeros(maxit,num_vec);
  res3 = zeros(maxit,num_vec);
end

% initialize iterates
R = B - A(X);
X = X + W*(WAW\(W'*R));
R = B - A(X);
R = R - W*(WW\(W'*R));  %<-- reorthogonalization
if ~isempty(M)
  Z = M(R);
else
  Z = R;
end
[P, gamma_inv] = qr(Z - W*(WAW\(AW'*Z)), 0);
RZ = R'*Z;

% the first residuals
for i = 1:num_vec
  res(1,i) = norm(R(:,i));
end
if exist('SOL', 'var')
  E = X - SOL;
  res2(1,:) = sqrt(abs(sum(E.*A(E))));
  res3(1,:) = sqrt(abs(sum(E.*E)))./normSOL;
end

% loop
iter = 1;
while iter < maxit & max(res(iter,:)) > tol

  % Q = A * P
  Q = A(P);

  % compute alpha
  alpha = pinv(gamma_inv' * (P'*Q)) * RZ;
  
  % update X
  X = X + P*alpha;
  
  % update R
  R = R - Q*alpha;
  R = R - W*(WW\(W'*R));  %<-- reorthogonalization

  % Z = M * R
  RZold = RZ;
  if ~isempty(M)
    Z = M(R);
  else
    Z = R;
  end
  RZ = R'*Z;
  
  % compute beta
  beta = gamma_inv * pinv(RZold) * RZ;
  
  % update P
  [P, gamma_inv] = qr(Z + P*beta - W*(WAW\(AW'*Z)), 0);
  
  % compute residuals
  iter = iter + 1;
  for i = 1:num_vec
    res(iter,i) = norm(R(:,i));
  end
  if exist('SOL', 'var')
    E = X - SOL;
    res2(iter,:) = sqrt(abs(sum(E.*A(E))));
    res3(iter,:) = sqrt(abs(sum(E.*E)))./normSOL;
  end

end

% scale back the computed solutions
for i = 1:num_vec
  X(:,i) = X(:,i) * normB(i);
end

% output simple message if fail
res = res(1:iter,:);
if exist('SOL', 'var')
  res2 = res2(1:iter,:);
  res3 = res3(1:iter,:);
end
if iter == maxit
  fprintf(1, '  DBPCG: maxit reached. res = %g\n', max(res(end,:)));
end
