function [eig_lo, n_lo, eig_hi, n_hi] = eigk(A, M, n, k, k1, k2, tol)
%function [eig_lo, n_lo, eig_hi, n_hi] = eigk(A, M, n, k, k1, k2, tol)
%
% This function computes the bottom k1 and top k2 eigenvalues of M*A,
% where both A and M are symmetric, using the Lanczos algorithm.
%
% INPUTS
%   A:      Matrix/Operator A, symmetric positive definite.
%   M:      Matrix/Operator M, symmetric positive definite. Can be empty.
%   n:      Dimension of A (and M).
%   k:      Maximum number of Lanczos steps.
%   k1:     Request k1 bottom eigenvalues.
%   k2:     Request k2 top eigenvalues.
%   tol:    Tolerance for checking convergence.
%
% OUTPUTS
%   eig_lo: The bottom Ritz values, sorted.
%   n_lo:   Number of bottom Ritz values (eigenvalues) that have converged.
%   eig_hi: The top Ritz values, sorted.
%   n_hi:   Number of top Ritz values (eigenvalues) that have converged.
%

%Author: J. Chen
%$Date: 2011/11/23 11:51:10$

u = randn(n,1);
[~, ~, T] = lan(A, M, n, u, k, k1, k2, tol);

[~, Theta] = eig(T);
[vecTheta, idx] = sort(diag(Theta));

vecTheta2 = eig(T(1:end-1,1:end-1));
vecTheta2 = sort(vecTheta2);

m = size(T,1) - 1;
eig_lo = vecTheta(1:min(m,k1));
n_lo = length(find( abs(eig_lo-vecTheta2(1:min(m,k1))) ./ ...
                    abs(eig_lo) < tol ));
eig_hi = vecTheta(end-min(m,k2)+1:end);
n_hi = length(find( abs(eig_hi-vecTheta2(end-min(m,k2)+1:end)) ./ ...
                    abs(eig_hi) < tol ));
