function [V, U, T, early] = lan(A, M, n, u, k, k1, k2, tol)
%function [V, U, T, early] = lan(A, M, n, u, k, k1, k2, tol)
%
% This function implements the Lanczos algorithm for (computing
% eigen-elements of) the matrix M*A, where both A and M are positive
% definite matrices.
%
%                 MA*V = V*T + ...;     V = MU
%
% INPUTS
%   A:     Matrix/Operator A, symmetric positive definite.
%   M:     Matrix/Operator M, symmetric positive definite. Can be empty.
%   n:     Dimension of A (and M).
%   u:     Initial vector input to the algorithm.
%   k:     Maximum number of Lanczos steps.
%   k1,k2: Terminate if bottom k1 and top k2 Ritz values converge.
%   tol:   Tolerance checking early termination.
%
% OUTPUTS
%   V, U:  Ritz vectors.
%   T:     Jacobi matrix.
%   early: Indicate if prematured termination happens. Can happen
%          when the requested Ritz values have converged, or norms
%          become negative.
%

%Author: J. Chen
%$Date: 2011/11/23 11:51:19$

if isnumeric(A)
  A = @(x)(A*x);
end
if ~isempty(M) & isnumeric(M)
  M = @(x)(M*x);
end

sigma  = zeros(k,1);
tau    = zeros(k+1,1);
V      = zeros(n,k+1);
U      = zeros(n,k+1);
T      = zeros(k+1);

if ~isempty(M)
  U(:,1) = u / sqrt(u'*M(u));
  V(:,1) = M(U(:,1));
else
  U(:,1) = u / norm(u);
  V(:,1) = U(:,1);
end
u_old  = zeros(n,1);

j = 1;
early = 0;
eig_lo = repmat(-inf,k1,1);
eig_hi = repmat(inf,k2,1);
while j <= k
  
  q = A(V(:,j)) - tau(j)*u_old;
  sigma(j) = q'*V(:,j);
  
  %--- early termination
  if sigma(j) < 0
    early = 1;
    break;
  end
  
  q = q - sigma(j)*U(:,j);

  %--- full reorthogonalization
  if ~isempty(M)
    q = q - U(:,1:j-1) * (M(q)'*U(:,1:j-1))';
  else
    q = q - U(:,1:j-1) * (q'*U(:,1:j-1))';
  end

  if ~isempty(M)
    Mq = M(q);
    tau(j+1) = sqrt(q'*Mq);
    U(:,j+1) = q/tau(j+1);
    V(:,j+1) = Mq/tau(j+1);
  else
    tau(j+1) = norm(q);
    U(:,j+1) = q/tau(j+1);
    V(:,j+1) = U(:,j+1);
  end
  
  T(j,j) = sigma(j);
  T(j,j+1) = tau(j+1);
  T(j+1,j) = tau(j+1);

  %--- early termination
  if j >= max([k1,k2])
    [S, Theta] = eig(T(1:j,1:j));
    [vecTheta, idx] = sort(diag(Theta));
    if max(abs(vecTheta(1:k1)-eig_lo)./abs(vecTheta(1:k1))) < tol ...
          & max(abs(vecTheta(end-k2+1:end)-eig_hi)./ ...
                abs(vecTheta(end-k2+1:end))) < tol
      early = 1;
      j = j + 1;
      break;
    end
  end
  
  u_old = U(:,j);
  j = j + 1;
  
end

V = V(:,1:j-1);
U = U(:,1:j-1);
T = T(1:j-1,1:j-1);
