function [score, iter] = score_func(K, dK, M, y, u, m, maxit, tol)
%function [score, iter] = score_func(K, dK, M, y, u, m, maxit, tol)
%
% This function implements the SAA of the score function (the gradient
% of the log-likelihood function). The Jacobian of the score equation
% is not computed.
%
% INPUTS
%   K, dK:         Covariance matrix and derivatives.
%   M:             Preconditioner. Can be empty.
%   y, u:          Appear in the SAA of the score function.
%   m, maxit, tol: Parameters for the linear solver.
%
% OUTPUTS
%   score:         Scores.
%   iter:          Numbers of iterations in the linear solver.
%

%Author: J. Chen
%$Date: 2011/12/19 18:22:37$

d = length(dK);

if isnumeric(K)
  K = @(x)(K*x);
end

if isnumeric(dK{1})
  for i = 1:d
    dK{i} = @(x)(dK{i}*x);
  end
end

if ~isempty(M) & isnumeric(M)
  M = @(x)(M*x);
end

N = size(u,2);

% solve K\u and K\y
B = [u y];
X0 = zeros(size(B));
n = size(B,1);
uu = randn(n,1);

if m == 0
  [Kuy, iter] = bpcg(K, B, M, X0, maxit, tol);
else
  W = lan(K, M, n, uu, m, m, m, tol);
  [Kuy, iter] = dbpcg(K, B, M, W, X0, maxit, tol);
end

% compute dK*u and dK*(K\y)
for j = 1:d
  dKu{j} = dK{j}(u);
  dKKy{j} = dK{j}(Kuy(:,N+1));
end

% assemble results
score = zeros(d,1);
for j = 1:d
  for i = 1:N
    score(j) = score(j) + Kuy(:,i)' * dKu{j}(:,i);
  end
  score(j) = score(j)/N - Kuy(:,N+1)' * dKKy{j};
end
