function [I, V, iter] = score_func_mle_covar_approx(K, dK, M, N, ...
                                                  u2, m, maxit, tol)
%function [I, V, iter] = score_func_mle_covar_approx(K, dK, M, N, ...
%                                                  u2, m, maxit, tol)
%
% This function returns the Fisher information matrix I, and the
% covariance matrix in the convergence result of the MLE estimate by
% using the SAA score equation approach. This function computes the
% matrices approximately. To compute the matrices exactly, use the
% function SCORE_FUNC_MLE_COVAR.
%
% INPUTS
%   K, dK:         Covariance matrix and derivatives.
%   M:             Preconditioner. Can be empty.
%   N:             Appear in the SAA of the score function.
%   u2:            Random vectors needed to approximate trace.
%   m, maxit, tol: Parameters for the linear solver.
%
% OUTPUTS
%   I:             Fisher information matrix.
%   V:             Covariance matrix in convergence result.
%   iter:          Numbers of iterations in the linear solver.
%

%Author: J. Chen
%$Date: 2011/12/19 17:49:14$

d = length(dK);

if isnumeric(K)
  K = @(x)(K*x);
end

if isnumeric(dK{1})
  for i = 1:d
    dK{i} = @(x)(dK{i}*x);
  end
end

if ~isempty(M) & isnumeric(M)
  M = @(x)(M*x);
end

n = size(u2,1);
N2 = size(u2,2);

% Wi*u = K\(dKi*u)
X0 = zeros(size(u2));
if m == 0
  for i = 1:d
    [Wu{i}, iter(i)] = bpcg(K, dK{i}(u2), M, X0, maxit, tol);
  end
else
  uu = randn(n,1);
  WW = lan(K, M, n, uu, m, m, m, tol);
  for i = 1:d
    [Wu{i}, iter(i)] = dbpcg(K, dK{i}(u2), M, WW, X0, maxit, tol);
  end
end

% Wi'*u = dKi*(K\u)
if m == 0
  [Ku, iter(d+1)] = bpcg(K, u2, M, X0, maxit, tol);
else
  [Ku, iter(d+1)] = dbpcg(K, u2, M, WW, X0, maxit, tol);
end
for i = 1:d
  WTu{i} = dK{i}(Ku);
end

% diag(Wi) \approx u.*(Wi*u)
for i = 1:d
  diagW{i} = zeros(n,1);
  for k = 1:N2
    diagW{i} = diagW{i} + Wu{i}(:,k) .* u2(:,k);
  end
  diagW{i} = diagW{i}/N2;
end

% I(i,j) = tr(Wi*Wj)/2 \approx (u'*Wi*Wj*u)/2
% J(i,j) = tr(Wi*Wj)/4 + tr(Wi*Wj')/4 - diag(Wi)'*diag(Wj)/2
I = zeros(d);
J = zeros(d);
for i = 1:d
  for j = 1:d
    I(i,j) = 0;
    for k = 1:N2
      I(i,j) = I(i,j) + WTu{i}(:,k)' * Wu{j}(:,k);
    end
    I(i,j) = I(i,j)/2/N2;
    
    J(i,j) = 0;
    for k = 1:N2
      J(i,j) = J(i,j) + WTu{j}(:,k)' * WTu{i}(:,k);
    end
    J(i,j) = J(i,j)/4/N2;
    J(i,j) = J(i,j) + I(i,j)/2 - diagW{i}'*diagW{j}/2;
  end
end
I = (I+I')/2;
J = (J+J')/2;

% V = inv(I) + inv(I)*J*inv(I)/N
V = (eye(d) + (I\J)/N) * inv(I);
V = (V+V')/2;
